# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import numpy as np

np.seterr(all="raise")

from quocslib.gradientfreemethods.DirectSearchMethod import DirectSearchMethod
from quocslib.stoppingcriteria.NelderMeadStoppingCriteria import (
    NelderMeadStoppingCriteria, )


class NelderMead(DirectSearchMethod):
    callback: callable

    def __init__(self, settings: dict, stopping_criteria: dict, callback: callable = None,
                 stop_optimization_callback: callable = None):
        """
        Nelder-Mead is an updating algorithm based on the simplex method.
        :param dict settings: settings for the NM algorithm
        :param dict stopping_criteria: stopping criteria
        """
        super().__init__()
        self.callback = callback
        # Active the parallelization for the firsts evaluations
        self.is_parallelized = settings.setdefault("parallelization", False)
        self.is_adaptive = settings.setdefault("is_adaptive", False)
        # TODO Create it using dynamical import module
        # Stopping criteria object
        stopping_criteria.setdefault("stop_function", stop_optimization_callback)
        self.sc_obj = NelderMeadStoppingCriteria(stopping_criteria)

    def run_dsm(self, func, x0, args=(), initial_simplex=None, max_eval_total=None, **kwargs) -> dict:
        """

        :param callable func: Function to be called at every function evaluation
        :param np.array x0: initial point
        :param tuple args: Further arguments
        :param np.array initial_simplex: Starting simplex for the Nelder Mead evaluation
        :param int max_eval_total: Maximum iteration number of function evaluations in total
        :return:
        """
        # Creation of the communication function for the OptimizationAlgorithm object
        calls_number, func = self._get_wrapper(args, func)
        # Set to false is_converged
        self.sc_obj.is_converged = False
        # update the total max of function evaluations
        self.sc_obj.max_eval_total = max_eval_total
        # Initialize the iteration number
        iterations = 0
        # Landscape dimension
        dim = len(x0)
        # Hyper-parameters for adaptive and not adaptive NM
        if self.is_adaptive:
            f_dim = float(dim)
            [rho, chi, psi, sigma] = [
                1,
                1 + 2 / f_dim,
                0.75 - 1 / (2 * f_dim),
                1 - 1 / f_dim,
            ]
        else:
            [rho, chi, psi, sigma] = [1, 2, 0.5, 0.5]
        # Start simplex initialization
        if initial_simplex is None:
            nonzdelt = 0.05
            zdelt = 0.00025
            sim = np.zeros((dim + 1, dim), dtype=x0.dtype)
            for k in range(dim):
                y = np.array(x0, copy=True)
                if y[k] != 0:
                    y[k] = (1 + nonzdelt) * y[k]
                else:
                    y[k] = zdelt
                sim[k + 1] = y
        else:
            sim = initial_simplex.copy()
        # Function evaluation array
        fsim = np.zeros((dim + 1, ), float)
        # Initial evaluation of the start simplex
        # TODO parallelization for start simplex initialization, i.e. send single data file for multiple evaluations!
        for k in range(dim + 1):
            fsim[k] = func(sim[k], iterations)
        # Sort the array by the lowest function value since we are performing a minimization
        ind = np.argsort(fsim)
        [sim, fsim] = [np.take(sim, ind, 0), np.take(fsim, ind, 0)]
        # Update NM function evaluation
        iterations = 1
        # Start loop for function evaluation
        while not self.sc_obj.is_converged:
            # Barycenter
            xbar = np.add.reduce(sim[:-1], 0) / dim
            # Reflection
            xr = (1 + rho) * xbar - rho * sim[-1]
            # 2nd reflection
            fxr = func(xr, iterations)
            # Shrinking
            doshrink = 0
            if fxr < fsim[0]:
                xe = (1 + rho * chi) * xbar - rho * chi * sim[-1]
                fxe = func(xe, iterations)
                if fxe < fxr:
                    sim[-1] = xe
                    fsim[-1] = fxe
                else:
                    sim[-1] = xr
                    fsim[-1] = fxr
            else:  # fsim[0] <= fxr
                if fxr < fsim[-2]:
                    sim[-1] = xr
                    fsim[-1] = fxr
                else:  # fxr >= fsim[-2]
                    # Perform contraction
                    if fxr < fsim[-1]:
                        xc = (1 + psi * rho) * xbar - psi * rho * sim[-1]
                        fxc = func(xc, iterations)
                        if fxc <= fxr:
                            sim[-1] = xc
                            fsim[-1] = fxc
                        else:
                            doshrink = 1
                    else:
                        # Perform an inside contraction
                        xcc = (1 - psi) * xbar + psi * sim[-1]
                        fxcc = func(xcc, iterations)

                        if fxcc < fsim[-1]:
                            sim[-1] = xcc
                            fsim[-1] = fxcc
                        else:
                            doshrink = 1
                    if doshrink:
                        for j in range(1, dim + 1):
                            sim[j] = sim[0] + sigma * (sim[j] - sim[0])
                            fsim[j] = func(sim[j], iterations)
            # Sort the array by the lowest function value since we are performing a minimization
            ind = np.argsort(fsim)
            [sim, fsim] = [np.take(sim, ind, 0), np.take(fsim, ind, 0)]
            # Increase the NM iteration
            iterations += 1
            # Update function evaluations number
            # self.sc_obj.fnc_evals = calls_number[0]
            # Check for error in the communication method
            if self.callback is not None:
                if not self.callback():
                    self.sc_obj.is_converged = True
                    self.sc_obj.terminate_reason = "User stopped the optimization"
            # Check stopping criteria
            self.sc_obj.check_stopping_criteria(sim, fsim, calls_number[0])
        # END of while loop
        # Fix the iteration number
        iterations = iterations - 1
        # Optimal parameters and value
        x = sim[0]
        fval = np.min(fsim)
        result_custom = {
            "F_min_val": fval,
            "X_opti_vec": x,
            "NitUsed": iterations,
            "NfunevalsUsed": calls_number[0],
            "terminate_reason": self.sc_obj.terminate_reason,
        }
        return result_custom



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from quocslib.optimizationalgorithms.OptimizationAlgorithm import OptimizationAlgorithm
from quocslib.Controls import Controls
from quocslib.gradientfreemethods.NelderMead import NelderMead
from quocslib.utils.dynamicimport import dynamic_import
from quocslib.tools.linearalgebra import simplex_creation


class DirectSearchAlgorithm(OptimizationAlgorithm):
    """ """

    initStatus = 0
    terminate_reason = "-1"

    def __init__(self, optimization_dict: dict = None, communication_obj=None):
        """
        :param optimization_dict:
        :param communication_obj:
        """
        super().__init__(communication_obj=communication_obj, optimization_dict=optimization_dict)
        ###########################################################################################
        # Direct Search method
        ###########################################################################################
        stopping_criteria = optimization_dict["algorithm_settings"]["dsm_settings"]["stopping_criteria"]
        # put global time limit into stopping_criteria so we don't have to pass it through functions
        optimization_dict["algorithm_settings"].setdefault("total_time_lim", 10**10)
        stopping_criteria.setdefault("total_time_lim", optimization_dict["algorithm_settings"]["total_time_lim"])
        direct_search_method_settings = optimization_dict["algorithm_settings"]["dsm_settings"]["general_settings"]
        if "dsm_name" in direct_search_method_settings:
            print("dsm_name is used direct search methods. This option is deprecated. Use \n"
                  "dsm_algorithm_module: quocslib.freegradients.NelderMead\n"
                  "dsm_algorithm_class: NelderMead")
            self.dsm_obj = NelderMead(direct_search_method_settings,
                                      stopping_criteria,
                                      callback=self.is_optimization_running)
        else:
            dsm_attribute = dynamic_import(
                module_name=direct_search_method_settings.setdefault("dsm_algorithm_module", None),
                class_name=direct_search_method_settings.setdefault("dsm_algorithm_class", None),
                name=direct_search_method_settings.setdefault("dsm_algorithm_name", None),
                class_type='dsm_settings')

            self.dsm_obj = dsm_attribute(direct_search_method_settings,
                                         stopping_criteria,
                                         callback=self.is_optimization_running)
        ###########################################################################################
        # Optimal algorithm variables ###########################################################
        ###########################################################################################
        # Starting FoM
        self.best_FoM = 1e10
        ###########################################################################################
        # Pulses, Parameters object ###########################################################
        ###########################################################################################
        self.controls = Controls(optimization_dict["pulses"], optimization_dict["times"],
                                 optimization_dict["parameters"])

    def _get_response_for_client(self):
        """
        Return True if a record is found
        Returns
        -------

        """
        is_record = False
        if self.FoM_dict["FoM"] < self.best_FoM:
            is_record = True
        response_dict = {"is_record": is_record, "FoM": self.FoM_dict["FoM"], "iteration_number": self.iteration_number}
        return response_dict

    def run(self):
        """

        Returns
        -------

        """
        # Direct search method
        self._dsm_build()

    def _dsm_build(self):
        """

        Returns
        -------

        """
        start_simplex = simplex_creation(self.controls.get_mean_value(), self.controls.get_sigma_variation())
        # Initial point for the Start Simplex
        x0 = self.controls.get_mean_value()
        # Initialize the best xx vector for this SI
        self.best_xx = self.controls.get_mean_value().copy()
        # Run the direct search algorithm
        result_l = self.dsm_obj.run_dsm(self._routine_call,
                                        x0,
                                        initial_simplex=start_simplex,
                                        sigma_v=self.controls.get_sigma_variation())
        # Update the results
        [self.best_FoM, self.xx,
         self.terminate_reason] = [result_l["F_min_val"], result_l["X_opti_vec"], result_l["terminate_reason"]]

    def _get_controls(self, xx):
        # pulses_list, time_grids_list, parameters_list
        [], [], parameters_list = self.controls.get_controls_lists(xx)

        controls_dict = {"pulses": [], "parameters": parameters_list, "timegrids": []}
        return controls_dict

    def _get_final_results(self):
        final_dict = {
            "Figure of merit": self.best_FoM,
            "parameters": self.xx,
            "terminate_reason": self.terminate_reason
        }
        return final_dict



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import numpy as np

from quocslib.pulses.basis.ChoppedBasis import ChoppedBasis
from quocslib.parameters.Parameter import Parameter
from quocslib.parameters.TimeParameter import TimeParameter
from quocslib.tools.randomgenerator import RandomNumberGenerator
from quocslib.utils.dynamicimport import dynamic_import


class Controls:
    """
    This is the main class for the optimization quantities, i.e. pulses, parameters, and times.
    All these quantities are defined in this class and can be accessed by calling the modules here.
    """
    def __init__(self, pulses_list, times_list, parameters_list, rng: RandomNumberGenerator = None):
        """
        Constructor of the general class containing all the controls used during the optimization

        :param pulses_list: List containing the settings for each pulse
        :param times_list: List containing the settings for each time
        :param parameters_list: List containing the setting for each parameter
        """
        # Map index
        map_index = -1
        # Lists and dictionary for the controls
        self.pulse_objs_list = []
        self.parameter_objs_list = []
        self.times_obj_dictionary = {}
        ###############################################
        # Pulses
        ###############################################
        self.pulses_number = len(pulses_list)
        for pulse in pulses_list:
            # Basis attribute inside the dictionary or module and class relative import
            basis = pulse["basis"]
            basis_attribute = dynamic_import(attribute=basis.setdefault("basis_attribute", None),
                                             module_name=basis.setdefault("basis_module", None),
                                             class_name=basis.setdefault("basis_class", None),
                                             name=basis.setdefault("basis_name", None),
                                             class_type='basis')
            # Create the pulse obj
            self.pulse_objs_list.append(basis_attribute(map_index, pulse, rng=rng))
            # Update the map index for the next control
            map_index = self.pulse_objs_list[-1].last_index
        ###############################################
        # Parameters
        ###############################################
        self.parameters_number = len(parameters_list)
        for parameter in parameters_list:
            self.parameter_objs_list.append(Parameter(map_index, parameter))
            # Update the map index for the next control
            map_index = self.parameter_objs_list[-1].last_index
        ###############################################
        # Times
        ###############################################
        self.times_number = len(times_list)
        for time in times_list:
            self.times_obj_dictionary[time["time_name"]] = TimeParameter(**time)
            # TODO Implement the time optimization here

    def get_control_parameters_number(self) -> int:
        """Return the control parameter number"""
        ###############################################
        # Pulses
        ###############################################
        control_parameters_number = 0
        for pulse_obj in self.pulse_objs_list:
            # Update number control parameters
            control_parameters_number += pulse_obj.control_parameters_number
        ###############################################
        # Parameters
        ###############################################
        for parameter_obj in self.parameter_objs_list:
            # Update number control parameters
            control_parameters_number += parameter_obj.control_parameters_number
        ###############################################
        # Times
        ###############################################
        # TODO Implement the time optimization here
        # Return the control parameters number
        return control_parameters_number

    def select_basis(self) -> None:
        """Initialize the superparameter basis"""
        for pulse in self.pulse_objs_list:
            pulse.super_parameter_distribution_obj.set_random_super_parameter()
            # Update the base pulse parameters and functions
            pulse.update_chopped_basis()
            # Update the control parameter indexes
            self._update_control_parameter_indexes()

    def _update_control_parameter_indexes(self) -> None:
        """Update the control parameter indexes"""
        # Map index
        map_index = -1
        ###############################################
        # Pulses
        ###############################################
        for pulse_obj in self.pulse_objs_list:
            pulse_obj.set_control_parameters_list(map_index)
            # Update the map index for the next control
            map_index = pulse_obj.last_index
        ###############################################
        # Parameters
        ###############################################
        for parameter_obj in self.parameter_objs_list:
            parameter_obj.set_control_parameters_list(map_index)
            # Update the map index for the next control
            map_index = parameter_obj.last_index
        ###############################################
        # Times
        ###############################################
        # TODO Implement the time optimization here
        # Return the control parameters number

    def get_random_super_parameter(self) -> np.array:
        """Return list with dcrab current super_parameters"""
        super_parameter_list = []
        for pulse in self.pulse_objs_list:
            if isinstance(pulse, ChoppedBasis):
                super_parameter_list.append(pulse.super_parameter_distribution_obj.w)
        super_parameter_array = np.asarray(super_parameter_list)
        return super_parameter_array

    def get_sigma_variation(self) -> np.array:
        """Return a numpy array with the maximum sigma in the parameters choice for the start simplex

        :return np.array
        """
        sigma_variation_coefficients = np.zeros(self.get_control_parameters_number(), dtype="float")
        # Pulses
        for pulse in self.pulse_objs_list:
            sigma_variation_coefficients[pulse.control_parameters_list] = pulse.scale_coefficients
        # Parameters
        for parameter in self.parameter_objs_list:
            sigma_variation_coefficients[parameter.control_parameters_list] = parameter.amplitude_variation
        # Times
        for time_name in self.times_obj_dictionary:
            time = self.times_obj_dictionary[time_name]
            if time.is_optimization:
                sigma_variation_coefficients[time.control_parameters_list] = time.amplitude_variation
        return sigma_variation_coefficients

    def get_mean_value(self) -> np.array:
        """Return a numpy array the mean value

        :return np.array:
        """
        mean_value_coefficients = np.zeros(self.get_control_parameters_number(), dtype="float")
        # Pulses
        for pulse in self.pulse_objs_list:
            mean_value_coefficients[pulse.control_parameters_list] = pulse.offset_coefficients
        # Parameters
        for parameter in self.parameter_objs_list:
            mean_value_coefficients[parameter.control_parameters_list] = parameter.value
        # Times
        # TODO In case of time optimization return the initial value
        return mean_value_coefficients

    def update_base_controls(self, optimized_parameters_vector: np.array) -> None:
        """Update the base controls. Only pulses is enough"""
        # Set the times
        for time_name in self.times_obj_dictionary:
            time = self.times_obj_dictionary[time_name]
            if time.is_optimization:
                time.set_parameter(optimized_parameters_vector[time.control_parameters_list])
        # Set the pulses
        for pulse in self.pulse_objs_list:
            time_name = pulse.time_name
            pulse.set_base_pulse(optimized_parameters_vector[pulse.control_parameters_list],
                                 final_time=self.times_obj_dictionary[time_name].get_time())
        # Set the parameters
        for parameter in self.parameter_objs_list:
            parameter.set_parameter(optimized_parameters_vector[parameter.control_parameters_list])

    def get_controls_lists(self, optimized_parameters_vector: np.array) -> [list, list, list]:
        """
        Set the optimized control parameters and get the controls

        :param np.array optimized_parameters_vector:
        :return: The pulses, time grids, and the parameters in three different lists of numpy arrays.
        """
        pulses_list = []
        time_grids_list = []
        parameters_list = []
        # Set the times
        for time_name in self.times_obj_dictionary:
            time = self.times_obj_dictionary[time_name]
            if time.is_optimization:
                time.set_parameter(optimized_parameters_vector[time.control_parameters_list])
        # Get the pulses and the timegrids
        for pulse in self.pulse_objs_list:
            time_name = pulse.time_name
            pulses_list.append(
                pulse.get_pulse(optimized_parameters_vector[pulse.control_parameters_list],
                                final_time=self.times_obj_dictionary[time_name].get_time()))
            time_grids_list.append(pulse.time_grid)
        # Get the parameters
        for parameter in self.parameter_objs_list:
            parameters_list.append(
                parameter.get_parameter(optimized_parameters_vector[parameter.control_parameters_list]))

        return pulses_list, time_grids_list, parameters_list



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
from abc import abstractmethod
import numpy as np


class SuperParameterDistribution:
    basis_vectors_number: int

    def __init__(self, lower_limit: float = 0.0, upper_limit: float = 5, **kwargs):
        # Create the array to store the basis vector super_parameters
        self.w = np.zeros((self.basis_vectors_number, ), dtype="float")
        self.lower_limit_w = lower_limit
        self.upper_limit_w = upper_limit

    @abstractmethod
    def set_random_super_parameter(self):
        """Spend some words here"""



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import numpy as np


class BaseParameter:
    """
    Parameter Class where every parameter to be optimized is defined. It is used as a parent class of the Time class in
    order to optimize the time of the pulse whenever the user specified it in the configuration class.
    """
    def __init__(
        self,
        map_index=-1,
        parameter_name="parameter",
        initial_value=1.0,
        lower_limit=0.1,
        upper_limit=1.1,
        amplitude_variation=0.1,
    ):
        """
        @param str parameter_name: Name of the parameter
        @param float initial_value: Initial value for the parameter to be used in the optimization
        @param float lower_limit: Lower limit for the parameter
        @param float upper_limit: Upper limit for the parameter
        @param float amplitude_variation: Amplitude variation for the simplex initialization
        """
        # Parameter name
        self.parameter_name = parameter_name
        # Initial value for the optimization
        self.value = initial_value
        # Upper limit
        self.upper_limit = upper_limit
        # Lower limit
        self.lower_limit = lower_limit
        # Amplitude variation
        self.amplitude_variation = amplitude_variation
        # The parameters control number is ever equal to 1
        self.control_parameters_number = 1
        # Create the control parameters list to get the optimized parameter from the optimized vector
        self.control_parameters_list = [map_index + i + 1 for i in range(self.control_parameters_number)]
        # Update the map_index number for the next pulse
        self.last_index = self.control_parameters_list[-1]

    def set_control_parameters_list(self, map_index):
        """Set the control parameters list. It is used when the"""
        self.control_parameters_list = [map_index + i + 1 for i in range(self.control_parameters_number)]

    def set_parameter(self, optimized_parameter_vector):
        """
        Set the parameter value after checking the constraints
        @param np.array optimized_parameter_vector: The optimized parameter coming from the optimization algorithm
        @return:
        """
        #
        self._set_parameter(optimized_parameter_vector[0])

    def get_parameter(self, optimized_parameter_vector):
        """

        @param optimized_parameter_vector:
        @return:
        """
        self._set_parameter(optimized_parameter_vector[0])
        return self.value

    def _set_parameter(self, parameter):
        """

        @param float parameter:
        @return:
        """
        self.value = self._check_limits(parameter)

    def _check_limits(self, parameter):
        """
        Check if the optimized parameter respect the amplitude limits before sending to the main controls
        @param float parameter: Parameter coming from the optimization
        @return float: Return the parameter after applying the constraints
        """
        #
        a = self.lower_limit
        #
        b = self.upper_limit
        return np.minimum(np.maximum(a, parameter), b)



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


class Wrap:
    def __init__(self, args):
        self.args = args
        self.calls_number = [0]
        self.function = None

    def function_wrapper(self, *wrapper_args):
        self.calls_number[0] += 1
        if len(self.args) == 0:
            return self.function(*wrapper_args)
        else:
            return self.function(*(wrapper_args + self.args))

    def wrap_function(self, function):
        if function is None:
            return self.calls_number, None
        self.function = function

        return self.calls_number, self.function_wrapper


class DirectSearchMethod:
    def __init__(self):
        # TODO Set the initial wrapper function with the target function and the optional arguments
        # TODO Integrate here the callback function for user interruption
        pass

    @staticmethod
    def _get_wrapper(args, func):
        wrap = Wrap(args)
        func = wrap.wrap_function(func)
        return func



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from abc import ABCMeta, abstractmethod
import os


class AbstractDump(metaclass=ABCMeta):
    """Abstract class for dumping data"""
    @abstractmethod
    def __init__(self, results_path: str = ".", **kwargs):
        """
        Abstract method for the constructor dumping classes. Set here the relevant paths
        :param str results_path: Path of the folder of the results
        """

    @abstractmethod
    def dump_controls(self,
                      pulses_list: list = [],
                      time_grids_list: list = [],
                      parameters_list: list = [],
                      **kwargs) -> None:
        """
        Abstract method for dumping the controls.
        :param list pulses_list: List of np.array. Every np.array is a pulse.
        :param list time_grids_list: List of np.array. Every np.array is a time grid at each time grid corresponds
        a pulse.
        :param list parameters_list: List of floats.
        :return:
        """



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import numpy as np
from scipy.optimize import minimize
import scipy

from quocslib.Controls import Controls
from quocslib.utils.dynamicimport import dynamic_import

from quocslib.timeevolution.piecewise_integrator import pw_evolution
from quocslib.tools.linearalgebra import commutator
from quocslib.pulses.basis.PiecewiseBasis import PiecewiseBasis


class GRAPEAlgorithm:
    """
    This is an implementation of the gradient ascent pulse engineering (GRAPE) algorithm for open-loop optimal control.
    The three important function are:
    * the constructor with the optimization dictionary and the communication object as parameters
    * run : The main loop for optimal control
    * _get_controls : return the set of controls as a dictionary with pulses, parameters, and times as keys
    * _get_final_results: return the final result of the optimization algorithm
    """
    def __init__(self, optimization_dict: dict = None):
        """
        This is the implementation of the GRAPE algorithm. All the arguments in the constructor are passed to the
        OptimizationAlgorithm class except the optimization dictionary where the GRAPE settings and the controls are defined.
        """
        ###########################################################################################
        # Optimal algorithm variables if any
        ###########################################################################################
        # Is empty
        self.alg_parameters = optimization_dict["algorithm_settings"]
        # Starting FoM
        self.best_FoM = 1e10
        ###########################################################################################
        # Pulses, Parameters, Times object
        ###########################################################################################
        # we will sculpt this a little since you have to be careful what you pass here
        # times are just the points we discretise at

        # might need to control if you change something

        self.A = optimization_dict["A"]
        self.B = optimization_dict["B"]
        self.n_slices = optimization_dict["n_slices"]
        self.rho_init = optimization_dict["rho_init"]
        self.rho_target = optimization_dict["rho_target"]
        self.dt = optimization_dict["dt"]
        self.sys_type = optimization_dict["sys_type"]
        self.dim = np.size(self.A, 1)
        self.num_pulses = len(self.B)
        self.initial_guess = optimization_dict["initial_guess"]
        self.FoM_list = []

        # create some storage arrays for the forward and backward propagated state
        self.rho_storage = np.array([self.rho_init for i in range(self.n_slices + 1)])
        self.rho_storage[0] = self.rho_init
        self.corho_storage = np.array([self.rho_target for i in range(self.n_slices + 1)])
        self.corho_storage[-1] = self.rho_target
        self.propagator_storage = np.array([self.A for i in range(self.n_slices)])

        self.iteration_number = None

        pw_basis_dict = {
            "pulse_name": "GRAPE",
            "bins_number": optimization_dict["n_slices"],
            "time_name": "",
            "lower_limit": -np.Inf,
            "upper_limit": np.Inf,
            "amplitude_variation": None,
            "initial_guess": {
                "function_type": "lambda_function",
                "lambda_function": "lambda x: self.initial_guess",
            },
            "scaling_function": {
                "function_type": "lambda_function",
                "lambda_function": "lambda x: x",
            },
        }
        # This declaration is wrong
        # pulse_dict = [
        #     {"basis": PiecewiseBasis(basis={}, **pw_basis_dict)}
        # ] * self.num_pulses
        pulse_dict = [{"basis": {"basis_attribute": PiecewiseBasis}, **pw_basis_dict}] * self.num_pulses
        time_dict = [{"time_name": ""}] * self.num_pulses
        param_dict = [{"parameter_name": ""}] * self.num_pulses

        # Initialize the control object
        self.controls = Controls(pulse_dict, time_dict, param_dict, rng=None)

    def functional(self, drive, A, B, n_slices, dt, U_store, rho_store, corho_store, sys_type):
        """Compute the fidelity functional for the defined problem

        :param np.array drive: this should be a flat array that will be resized into N_ctrls x N_slices
        :param np.matrix A: drift Hamiltonian
        :param List[np.matrix] B: control Hamiltonians in a list of N_ctrls long
        :param int n_slices: the number of pulse slices
        :param float dt: the duration of each timeslice
        :param List[np.matrix] U_store: a store for all the propagators
        :param List[np.matrix] rho_store: a store for all the forward propagated states
        :param List[np.matrix] corho_store: a store for all the reverse propagated states
        :param str sys_type: either specifying state transfer or other
        :return Tuple[float, np.array]: Returns a tuple containing the gradient and the figure of merit
        """
        K = self.num_pulses
        drive = drive.reshape((K, n_slices))
        # TODO: Make this statement more clear
        # Update the propagator_storage class variable
        pw_evolution(U_store, drive, A, B, n_slices, dt)

        for t in range(n_slices):
            U = U_store[t]
            # depending on system type do different evolution
            if sys_type == "StateTransfer":
                rho_store[t + 1] = U @ rho_store[t] @ U.T.conj()
            else:
                rho_store[t + 1] = U @ rho_store[t]
        for t in reversed(range(n_slices)):
            U = U_store[t]
            if sys_type == "StateTransfer":
                corho_store[t] = U.T.conj() @ corho_store[t + 1] @ U
            else:
                corho_store[t] = corho_store[t + 1] @ U.T.conj()

        # then compute the gradients
        grads = np.zeros((K, n_slices))
        for k in range(K):
            for t in range(n_slices):
                if sys_type == "StateTransfer":
                    g = (1j * dt * corho_store[t].T.conj() @ commutator(B[k], rho_store[t]))
                    grads[k, t] = np.real(np.trace(g))
                else:
                    grads[k, t] = 0.0
        grads = grads.flatten()
        if sys_type == "StateTransfer":
            fid = 1 - np.abs(np.trace(corho_store[-1].T.conj() @ rho_store[-1]))
        else:
            fid = 0.0

        self.FoM_list.append(fid)

        return fid, grads

    def _get_functional(self):
        """Generates a lambda x: where x is the control

        :return lambda:
        """
        return lambda x: self.functional(
            x,
            self.A,
            self.B,
            self.n_slices,
            self.dt,
            self.propagator_storage,
            self.rho_storage,
            self.corho_storage,
            self.sys_type,
        )

    def run(self, init) -> None:
        """Main loop of the optimization"""

        func_topt = self._get_functional()
        # now we can optimize
        # need to be able to include things
        oo = scipy.optimize.minimize(func_topt, init, method="L-BFGS-B", jac=True, options=self.alg_parameters)

        # need to be able to implement pulses in Marco's way, ask him later
        self.best_FoM = oo.fun
        self.optimized_pulses = oo.x  # TODO we might want to reshape this
        self.opt_res = oo
        self.iteration_number = oo.nfev

    def _get_controls(self, xx: np.array) -> dict:
        """Get the controls dictionary from the optimized control parameters"""
        # [pulses, timegrids, parameters] = self.controls.get_controls_lists(xx)
        pulses = [self.optimized_pulses]
        timegrids = [np.ones(self.n_slices) * self.dt]
        parameters = []
        #
        controls_dict = {
            "pulses": pulses,
            "parameters": parameters,
            "timegrids": timegrids,
        }
        return controls_dict

    def _get_final_results(self) -> dict:
        """Return a dictionary with final results to put into a dictionary"""
        final_dict = {
            "FoM": self.best_FoM,
            "nfev": self.iteration_number,
        }
        return final_dict



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
import numpy as np

from quocslib.optimizationalgorithms.OptimizationAlgorithm import OptimizationAlgorithm
from quocslib.Controls import Controls
from quocslib.utils.dynamicimport import dynamic_import

from quocslib.tools.linearalgebra import simplex_creation


class AlgorithmTemplate(OptimizationAlgorithm):
    """
    This is the template for an algorithm class. The three important function are:
    * the constructor with the optimization dictionary and the communication object as parameters
    * run : The main loop for optimal control
    * _get_response_for_client : return info about the goodness of the controls and errors if any
    * _get_controls : return the set of controls as a dictionary with pulses, parameters, and times as keys
    * _get_final_results: return the final result of the optimization algorithm
    """
    def __init__(self, optimization_dict: dict = None, communication_obj=None):
        """
        This is the implementation of the dCRAB algorithm. All the arguments in the constructor are passed to the
        OptimizationAlgorithm class except the optimization dictionary where the dCRAB settings and the controls are defined.
        """
        super().__init__(communication_obj=communication_obj)
        ###########################################################################################
        # Inner free gradient method
        ###########################################################################################
        stopping_criteria = optimization_dict["algorithm_settings"]["dsm_settings"]["stopping_criteria"]
        # put global time limit into stopping_criteria so we don't have to pass it through functions
        optimization_dict["algorithm_settings"].setdefault("total_time_lim", 10**10)
        stopping_criteria.setdefault("total_time_lim", optimization_dict["algorithm_settings"]["total_time_lim"])
        direct_search_method_settings = optimization_dict["algorithm_settings"]["dsm_settings"]["general_settings"]
        dsm_attribute = dynamic_import(
            class_name="GradientFreeTemplate",
            module_name="quocslib.gradientfreemethods.GradientFreeTemplate",
        )
        self.dsm_obj = dsm_attribute(
            direct_search_method_settings,
            stopping_criteria,
            callback=self.is_optimization_running,
        )
        ###########################################################################################
        # Optimal algorithm variables if any
        ###########################################################################################
        alg_parameters = optimization_dict["algorithm_settings"]
        # Starting FoM
        self.best_FoM = 1e10
        ###########################################################################################
        # Pulses, Parameters, Times object
        ###########################################################################################
        # Initialize the control object
        self.controls = Controls(
            optimization_dict["pulses"],
            optimization_dict["times"],
            optimization_dict["parameters"],
        )

    def _get_response_for_client(self) -> dict:
        """Return useful information for th interface"""
        is_record = False
        FoM = self.FoM_dict["FoM"]
        if FoM < self.best_FoM:
            self.best_FoM = FoM
            is_record = True
        response_dict = {
            "is_record": is_record,
            "FoM": FoM,
            "iteration_number": self.iteration_number,
        }
        return response_dict

    def run(self) -> None:
        """Main loop of the optimization"""
        for super_it in range(1, 2):
            # Check if the optimization was stopped by the user
            if not self.is_optimization_running():
                return
            # Initialize the random super_parameters
            self.controls.select_basis()
            # Direct search method
            if super_it == 1:
                self._dsm_build(self.max_num_function_ev)
            else:
                self._dsm_build(self.max_num_function_ev2)
            # Update the base current pulses
            self._update_base_pulses()

    def _update_base_pulses(self) -> None:
        """Update the base dCRAB pulse"""
        self.controls.update_base_controls(self.xx)

    def _dsm_build(self, max_iteration_number: int) -> None:
        """Build the direct search method and run it"""
        start_simplex = simplex_creation(self.controls.get_mean_value(), self.controls.get_sigma_variation())
        # Initial point for the Start Simplex
        x0 = self.controls.get_mean_value()
        # Run the direct search algorithm
        result_l = self.dsm_obj.run_dsm(self._routine_call,
                                        x0,
                                        initial_simplex=start_simplex,
                                        max_eval_total=max_iteration_number)
        # Update the results
        [FoM, self.xx, self.terminate_reason] = [
            result_l["F_min_val"],
            result_l["X_opti_vec"],
            result_l["terminate_reason"],
        ]

    def _get_controls(self, xx: np.array) -> dict:
        """Get the controls dictionary from the optimized control parameters"""
        [pulses, timegrids, parameters] = self.controls.get_controls_lists(xx)
        #
        controls_dict = {
            "pulses": pulses,
            "parameters": parameters,
            "timegrids": timegrids,
        }
        return controls_dict

    def _get_final_results(self) -> dict:
        """Return a dictionary with final results to put into a dictionary"""
        final_dict = {
            "Figure of merit": self.best_FoM,
            "total number of function evaluations": self.iteration_number,
        }
        return final_dict



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
from abc import abstractmethod
import numpy as np
from quocslib.communication.AllInOneCommunication import AllInOneCommunication
from quocslib import __VERSION__ as QUOCSLIB_VERSION

INITIAL_FOM: float = 10**10


class OptimizationAlgorithm:
    init_status: bool = False
    FoM_maximum: float = 1e10
    xx: np.array
    alg_iteration_number: int
    iteration_number: int
    FoM_dict: dict

    def __init__(self, communication_obj: AllInOneCommunication = None, optimization_dict: dict = None):
        """
        The constructor of the OptimizationAlgorithm class. All the algorithms has to inherit it. It provides all the basic
        modules an optimizer should have. All the arguments are passed to the communication object. Find all the info
        in that class.

        :param dict communication_obj: Object fo the communication class
        """
        self.comm_obj = communication_obj
        # Print optimization dictionary into a file
        self.comm_obj.print_optimization_dictionary(optimization_dict)
        # Initialize the total iteration number, i.e. the total function evaluations of the algorithm
        self.iteration_number = 0
        # Update status
        self.init_status = True
        # Random number generator
        self.rng = None
        # Maximization or minimization
        # optimization_direction
        optimization_direction = optimization_dict.setdefault("optimization_direction", "minimization")
        if optimization_direction == "minimization":
            self.optimization_factor = -1.0
        elif optimization_direction == "maximization":
            self.optimization_factor = 1.0
        else:
            message = "You can choose between maximization/minimization " \
                      "only, but {0} is provided".format(optimization_direction)
            self.comm_obj.print_logger(message=message, level=40)
            raise TypeError
        self.best_FoM = self.optimization_factor * (-1.0) * INITIAL_FOM
        message = "The optimization direction is {0}".format(optimization_direction)
        self.comm_obj.print_logger(message=message, level=20)

    def begin(self) -> None:
        """Initialize the communication with the client"""
        # Open the log with the QuOCS version number
        self.comm_obj.print_logger("QuOCS version number: {0}".format(QUOCSLIB_VERSION))
        # Send starting message to the interface
        self.comm_obj.send_message("start")
        # Assign new job number to the client
        self.comm_obj.assign_job()
        # Send the new data for the communication to the client
        self.comm_obj.send_communication_data()
        # Notify it to the client
        self.comm_obj.update_init_msg_server(upd_name=self.comm_obj.client_job_name)

    def _routine_call(self, optimized_control_parameters: np.array, iterations: int) -> float:
        """
        General routine for any control algorithm. It has to be given as the argument of the inner free gradient control
        methods

        :param: np.array optimized_control_parameters: The vector with all the optimized control parameters
        :param: int iterations: Iteration number of the inner free gradient method
        :return: float: Return the figure of merit to the inner free gradient method
        """
        # Check if the optimization is still running
        is_running = self.comm_obj.get_user_running()
        if not is_running:
            return self.FoM_maximum
        # Update parameter array and iteration number
        self.xx, self.alg_iteration_number = optimized_control_parameters, iterations
        # Update iteration number
        self.iteration_number += 1
        # General workflow
        # Check interface update
        self.comm_obj.check_msg_client()
        # Send the controls to the interface
        self.comm_obj.send_controls(self._get_controls(optimized_control_parameters))
        # Update the notification file for the interface
        self.comm_obj.update_msg_server()
        # The interface reads the controls, calculates the FoM and updates its notification
        # Check for interface response
        self.comm_obj.check_msg_client()
        # Check if the optimization is still running
        is_running = self.comm_obj.get_user_running()
        if not is_running:
            return self.FoM_maximum
        # Get the figure of merit and update it to the main algorithm
        self.FoM_dict = self.comm_obj.get_data()["FoM_values"]
        # Send the response for the interface
        self.comm_obj.send_FoM_response(self._get_response_for_client())
        # Update the notification file for the interface
        self.comm_obj.update_msg_server()
        # The interface reads the FoM response and update its notification file
        #
        # Return the figure of merit, i.e. a real number, to the optimal based algorithm
        return -1.0 * self.optimization_factor * self.FoM_dict["FoM"]

    def get_is_record(self, FoM: float) -> bool:
        """Check if the figure of merit provided is a new record

        :param: FoM  : figure of merit provided by the user
        """
        # Minimization
        if self.optimization_factor < 0.0:
            if FoM < self.best_FoM:
                return True
        else:
            # Maximization
            if FoM > self.best_FoM:
                return True
        return False

    @abstractmethod
    def run(self) -> None:
        """Run the optimization algorithm"""
        raise NotImplementedError("Must override method in the Optimal Algorithm class")

    @abstractmethod
    def _get_response_for_client(self) -> dict:
        """Return a dictionary with useful info for the client interface. At least the dictionary
        has to provide "is_record": bool and "FoM": float"""
        raise NotImplementedError("Must override method in the Optimal Algorithm class")

    @abstractmethod
    def _get_controls(self, optimized_control_parameters: np.array) -> [list, list, list]:
        """Given the optimized control parameters, the control object in the optimal algorithm builds the
        pulses, time grids, and parameters"""
        raise NotImplementedError("Must override method in the Optimal Algorithm class")

    @abstractmethod
    def _get_final_results(self) -> dict:
        """The optimal algorithm gives back a dictionary with useful results"""
        raise NotImplementedError("Must override method in the Optimal Algorithm class")

    def is_optimization_running(self) -> bool:
        """Function to check if the optimization is still running"""
        return self.comm_obj.get_user_running()

    def stop_optimization(self) -> None:
        """Function to stop the optimization (inner direct search algorithm)"""
        self.comm_obj.set_is_running_state(value=False)

    def end(self) -> None:
        """Finalize the transmission with  the client"""
        # Check client update
        self.comm_obj.check_msg_client()
        # End communication
        self.comm_obj.end_communication(self._get_final_results())
        # Update server message
        self.comm_obj.update_msg_server()



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from abc import ABCMeta, abstractmethod


class AbstractFoM(metaclass=ABCMeta):
    """Abstract class for figure of merit evaluation"""
    @abstractmethod
    def get_FoM(self, pulses_list: list = [], time_grids_list: list = [], parameters_list: list = []) -> dict:
        """
        Abstract method for figure of merit evaluation. It returns a dictionary with
         the FoM key inside
        :param list pulses_list: List of np.array. Every np.array is a pulse.
        :param list time_grids_list: List of np.array. Every np.array is a time grid at each time grid corresponds
        to a pulse.
        :param list parameters_list: List of floats.
        :return dict: The dictionary must contain at least the "FoM" key with the figure of merit float. Other possible
        keys can provide useful information for errors or other.
        """



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import json
import numpy as np
import inspect


def readjson(filename: str) -> [int, dict]:
    """
    Read a json file given its filename
    :param str filename: Filename of the json file
    :return list error_code, user dictionary : The error code is 1 in case an error occurs.
    """
    err_stat = 0
    user_data = None
    try:
        with open(filename, "r") as file:
            user_data = json.load(file)
    except json.decoder.JSONDecodeError:
        err_stat = 1
        print('\n!!! The json file \"' + filename + '\" is not formatted properly.')
    except Exception as ex:
        print(ex)
        err_stat = 1
        print('\n!!! The json file \"' + filename + '\" was not found\n'
                                                    'or some other error occured while reading the file.')
    finally:
        return user_data


class ObjectEncoder(json.JSONEncoder):
    """Convert numpy array to list"""
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if obj is callable:
            return None
        if inspect.isclass(obj):
            return None
        return json.JSONEncoder.default(self, obj)


def writejsonfile(json_file: str, kwargs_bib: dict) -> int:
    """A wrapper to the json dump file with error code as return"""
    err_stat = 0
    try:
        with open(json_file, "w") as fp:
            json.dump(kwargs_bib, fp, indent=4, cls=ObjectEncoder)
    except Exception as ex:
        print("It is not possible to write the json file")
        err_stat = 1
    finally:
        return err_stat



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
from quocslib.utils.dynamicimport import dynamic_import
from quocslib.pulses.BasePulse import BasePulse
from quocslib.tools.randomgenerator import RandomNumberGenerator


class ChoppedBasis(BasePulse):
    """
    General class for chopped basis. All the chopped basis has to inherit this class.
    """

    super_parameter_number: int

    def __init__(self, basis: dict = None, rng: RandomNumberGenerator = None, **kwargs):
        """

        :param basis:
        :param kwargs:
        """
        super().__init__(**kwargs)
        super_parameter_distribution_dict = basis["random_super_parameter_distribution"]
        # Distribution attribute
        distribution_attribute = dynamic_import(
            attribute=super_parameter_distribution_dict.setdefault("distribution_attribute", None),
            module_name=super_parameter_distribution_dict.setdefault("distribution_module", None),
            class_name=super_parameter_distribution_dict.setdefault("distribution_class", None),
            name=super_parameter_distribution_dict.setdefault("distribution_name", None),
            class_type='superparameter_distribution')
        self.super_parameter_distribution_obj = distribution_attribute(self.super_parameter_number,
                                                                       super_parameter_distribution_dict,
                                                                       rng=self.rng)

    # Implement here other modules for Chopped Random Basis

    def update_chopped_basis(self) -> None:
        """Update chopped basis parameter. This function has to be implemented in the Basis class in case it needs, and
        used in the algorithm whenever at the begin of the new super iteration
        """



# a collection of potentially useful pulse tools
from scipy import interpolate
import matplotlib.pyplot as plt
import numpy as np


def visualise_pulse(dt, n_slices, optimised_pulse, n_pulses):
    # this can only run if we have something in the optimised_pulse really

    tArray = np.array([dt] * n_slices).cumsum() - dt

    f, ax = plt.subplots()
    for i in n_pulses:
        ax.bar(tArray, optimised_pulse[:, i], label=str(i), width=dt / 2)
    ax.set_xlabel("Time")
    ax.legend()
    plt.show()
    return f


def resample_pulse(dt, n_slices, optimised_pulse, n_pulses, time_axis):
    # interpolate pulse onto time axis given by
    ret = np.zeros((len(time_axis), n_pulses))
    original_time = np.array([dt] * n_slices).cumsum() - dt
    for i in range(n_pulses):
        e0 = optimised_pulse[0, i]
        e1 = optimised_pulse[-1, i]
        f = interpolate.interp1d(
            original_time,
            optimised_pulse[:, i],
            kind="nearest",
            fill_value=(e0, e1),
            bounds_error=False,
        )
        ret[:, i] = f(time_axis)
    return ret



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from quocslib.handleexit.HandleExit import HandleExit
from quocslib.utils.dynamicimport import dynamic_import
from quocslib.communication.AllInOneCommunication import AllInOneCommunication
from quocslib.utils.BestDump import BestDump
from quocslib.utils.AbstractFoM import AbstractFoM


class Optimizer:
    def __init__(self, optimization_dict: dict = None, FoM_object: AbstractFoM = None):
        """
        Write this docstring
        """
        self.interface_job_name = optimization_dict.setdefault("optimization_client_name", "run")
        self.communication_obj = AllInOneCommunication(interface_job_name=self.interface_job_name,
                                                       FoM_obj=FoM_object,
                                                       handle_exit_obj=HandleExit(),
                                                       dump_attribute=BestDump)

        self.results_path = self.communication_obj.results_path

        algorithm_dict = optimization_dict['algorithm_settings']
        self.optimizer_attribute = dynamic_import(attribute=algorithm_dict.setdefault("algorithm_attribute", None),
                                                  module_name=algorithm_dict.setdefault("algorithm_module", None),
                                                  class_name=algorithm_dict.setdefault("algorithm_class", None),
                                                  name=algorithm_dict.setdefault("algorithm_name", None),
                                                  class_type='algorithm')

        self.opt_alg_obj = self.optimizer_attribute(optimization_dict=optimization_dict,
                                                    communication_obj=self.communication_obj)

    def execute(self):
        """ Write this docstring """
        self.opt_alg_obj.begin()
        self.opt_alg_obj.run()
        self.opt_alg_obj.end()



import numpy as np
from scipy.linalg import expm

# can we do conditional import?
# try:
#     import jax.scipy as jsp
# except:
#     raise ImportError


# TODO Shall we merge the two functions in one ???
def pw_evolution(U_store, drive, A, B, n_slices, dt):
    """Compute the piecewise evolution of a system defined by the
    Hamiltonian H = A + drive * B and store the result in U_store

    :param List[np.matrix] U_store: the storage for all of the computed propagators
    :param np.array drive: an array of dimension n_controls x n_slices that contains the amplitudes of the pulse
    :param np.matrix A: the drift Hamiltonian
    :param List[np.matrix] B: the control Hamiltonians
    :param int n_slices: number of slices
    :param float dt: the duration of each time slice
    :return None: Stores the new propagators so this doesn't return
    """
    K = len(B)
    for i in range(n_slices):
        H = A
        for k in range(K):
            H = H + drive[k, i] * B[k]
        U_store[i] = expm(-1j * dt * H)
    return None


def pw_final_evolution(drive, A, B, n_slices, dt, u0):
    """Compute the piecewise evolution of a system defined by the
    Hamiltonian H = A + drive * B and concatenate all the propagators

    :param List[np.matrix] U_store: the storage for all of the computed propagators
    :param np.array drive: an array of dimension n_controls x n_slices that contains the amplitudes of the pulse
    :param np.matrix A: the drift Hamiltonian
    :param List[np.matrix] B: the control Hamiltonians
    :param int n_slices: number of slices
    :param np.matrix u0: the initial density matrix to start from
    :return np.matrix: the final propagator
    """
    K = len(B)
    U = u0
    for i in range(n_slices):
        H = A
        for k in range(K):
            H = H + drive[k, i] * B[k]
        U = expm(-1j * dt * H) @ U
    return U



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np


# TODO Add the Hamiltonians and spin operators as constant variables, in the python sense obviously
def hamiltonian_d1_d2(ft, delta1=0.0, delta2=0.0):
    sigma_x = _get_sigma_x()
    sigma_z = _get_sigma_z()

    ham_t = delta1 * sigma_z / 2 + (ft + delta2) * sigma_x / 2
    return ham_t


def hamiltonian_d1_d2_2fields(amplitude_t, phase_t, delta1=0.0, delta2=0.0):
    sigma_x = _get_sigma_x()
    sigma_y = _get_sigma_y()
    sigma_z = _get_sigma_z()
    # The controls
    ham_t = delta1 * sigma_z / 2 + amplitude_t * (1 + delta2) * (np.cos(phase_t) * sigma_x + np.sin(phase_t) * sigma_y)
    return ham_t


def _get_sigma_x():
    sigma_x = np.array([[0, 1], [1, 0]], dtype="complex")
    return sigma_x


def _get_sigma_y():
    sigma_y = np.array([[0, -1j], [1j, 0]], dtype="complex")
    return sigma_y


def _get_sigma_z():
    sigma_z = np.array([[1, 0], [0, -1]], dtype="complex")
    return sigma_z



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import numpy as np

from quocslib.tools.randomgenerator import RandomNumberGenerator
from quocslib.pulses.BasePulse import BasePulse


class PiecewiseBasis(BasePulse):
    amplitude_variation: float
    optimized_control_parameters: np.ndarray
    optimized_super_parameters: np.ndarray
    time_grid: np.ndarray

    def __init__(self, map_index, pulse_dictionary: dict, rng: RandomNumberGenerator = None):
        """

        :param int map_index: Index number to use to get the control parameters for the Fourier basis
        :param dict pulse_dictionary: The dictionary of the pulse defined here. Only the basis dictionary is used btw

        :param dict pulse_dictionary: Should contain the basis_dict under the key "basis" and should contain the number of bins and time spacing under "n_bins" and "dt"
        """
        # basis_dict = pulse_dictionary["basis"]
        # Constructor of the parent class, i.e. Chopped Basis
        self.control_parameters_number = pulse_dictionary["bins_number"]
        super().__init__(map_index=map_index, rng=rng, **pulse_dictionary)
        # self.control_parameters_number = 1
        # super().__init__(**kwargs)
        # basis_attribute = dynamic_import(attribute=basis.setdefault("basis_attribute", None),
        #                                     module_name=basis.setdefault("basis_module", None),
        #                                     class_name=basis.setdefault("basis_class", None))

        #################
        # Basis dependent settings
        #################

    def setdefault(a, b, c):
        class Skipper:
            def __init__(self):
                self.last_index = 0

        return lambda x, y: Skipper()

    # TODO check if this function is really called by GRAPE
    def _get_shaped_pulse(self) -> np.array:
        """Definition of the pulse parametrization. It is called at every function evaluation to build the pulse"""
        #################
        # Standard Basis Settings: amplitude limits, amplitude variation for the simplex,
        # distribution of super parameters, etc ...
        ################
        pulse = self.pulse_amplitudes
        # in the case of a piecewise basis the pulse is defined by the amplitudes

        return pulse



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
from abc import abstractmethod
import numpy as np
from datetime import datetime


class StoppingCriteria:

    def __init__(self, stopping_criteria: dict = None):
        """
        Parent class for the stopping criteria
        """
        self.max_eval_total = None
        self.total_func_eval_track = 0
        self.temporary_func_eval_track = 0
        self.max_eval = stopping_criteria.setdefault("max_eval", 10**10)
        self.FoM_goal = stopping_criteria.setdefault("FoM_goal", -10**10)
        self.total_time_lim = stopping_criteria.setdefault("total_time_lim", 10**10)
        self.time_lim = stopping_criteria.setdefault("time_lim", 10**10)
        self.direct_search_start_time = datetime.now()
        self.total_start_time = datetime.now()
        self.stop_opt_function = stopping_criteria.setdefault("stop_function", None)
        pass

    @abstractmethod
    def check_stopping_criteria(self, **kwargs) -> None:
        """
        :return:
        """
        raise ValueError("Must be implemented in the custom stopping criteria")

    # TODO Implement handle exit check
    def check_user_stop(self):
        if False:
            self.is_converged = True

    def check_func_eval_single_direct_search(self, func_evaluations_single_direct_search: int) -> [bool, str]:
        """
        Check whether the maximum number of function evaluations has been exceeded

        :param float func_evaluations_single_direct_search: number of current function evaluation
        :return bool is_converged: True if the stopping criterion is fulfilled
        :return str terminate_reason: reason for the terminatiom
        """
        terminate_reason = "Exceeded number of allowed function evaluations per direct search."
        is_converged = False
        if func_evaluations_single_direct_search >= self.max_eval:
            is_converged = True
        return [is_converged, terminate_reason]

    def check_func_eval_total(self, func_evaluations_single_direct_search: int) -> [bool, str]:
        """
        Check whether the total maximum number of function evaluations has been exceeded

        :param float func_evaluations_single_direct_search: number of current function evaluation
        :return bool is_converged: True if the stopping criterion is fulfilled
        :return str terminate_reason: reason for the terminatiom
        """

        # ToDo: This is extremely complicated and ther has to ba a simple way to do this properly.
        #  Also this might break for very low max_eval and max_eval_total
        #  AND this actually breaks if we continue after reaching max_eval_total !!!
        #  But we don't reach that because we stop the optimization anyway

        if self.max_eval_total is not None:

            local_func_eval_track = 0

            if self.temporary_func_eval_track > func_evaluations_single_direct_search:
                self.total_func_eval_track += self.temporary_func_eval_track
                self.temporary_func_eval_track = 0

            local_func_eval_track = self.total_func_eval_track + func_evaluations_single_direct_search
            self.temporary_func_eval_track = func_evaluations_single_direct_search

            terminate_reason = "Exceeded number of allowed function evaluations in total."
            is_converged = False

            if local_func_eval_track >= self.max_eval_total:
                is_converged = True
                # this is a bit confusing but here we set is_running to False if is is linked
                if self.stop_opt_function is not None:
                    self.stop_opt_function()
            return [is_converged, terminate_reason]

        else:
            terminate_reason = "Exceeded number of allowed function evaluations in total."
            is_converged = False
            return [is_converged, terminate_reason]



    def check_simplex_criterion(self, sim: np.array) -> [bool, str]:
        """
        Check whether the simplex has converged

        :param np.array sim: current simplex
        :return bool is_converged: True if the stopping criterion is fulfilled
        :return str terminate_reason: reason for the terminatiom
        """
        terminate_reason = "Convergence of the simplex."
        is_converged = False
        if np.max(np.ravel(np.abs(sim[1:] - sim[0]))) <= self.xatol:
            is_converged = True
        return [is_converged, terminate_reason]

    def check_f_size(self, fsim: np.array) -> [bool, str]:
        """
        Check whether the FoM has converged inside this simplex

        :param np.array fsim: FoM values for current simplex
        :return bool is_converged: True if the stopping criterion is fulfilled
        :return str terminate_reason: reason for the terminatiom
        """
        terminate_reason = "Convergence of the FoM."
        is_converged = False
        try:
            maxDeltaFoMRel = np.max(np.abs(fsim[0] - fsim[1:])) / (np.abs(fsim[0]))
        except (ZeroDivisionError, FloatingPointError):
            maxDeltaFoMRel = fsim[1]
        if maxDeltaFoMRel <= self.frtol:
            is_converged = True
        return [is_converged, terminate_reason]

    def check_goal_reached(self, FoM: float) -> [bool, str]:
        """
        Check whether the FoM goal has been reached

        :param float FoM: current FoM
        :return bool is_converged: True if the stopping criterion is fulfilled
        :return str terminate_reason: reason for the terminatiom
        """
        terminate_reason = "FoM goal reached."
        is_converged = False
        if FoM <= self.FoM_goal:
            is_converged = True
            # this is a bit confusing but here we set is_running to False if is is linked
            if self.stop_opt_function is not None:
                self.stop_opt_function()
        return [is_converged, terminate_reason]

    def reset_direct_search_start_time(self) -> None:
        """
        Reset the direct_search_start_time for a new SI
        """
        self.direct_search_start_time = datetime.now()

    def check_direct_search_time_out(self) -> [bool, str]:
        """
        Check whether the optimization in a direct search has been running for too long.

        :return bool is_converged: True if the stopping criterion is fulfilled
        :return str terminate_reason: reason for the terminatiom
        """
        end_time = datetime.now()
        minutes_diff = (end_time - self.direct_search_start_time).total_seconds() / 60.0
        terminate_reason = "Optimization time for direct search exceeded limit."
        is_converged = False
        if minutes_diff >= self.time_lim:
            is_converged = True
        return [is_converged, terminate_reason]

    def check_total_time_out(self) -> [bool, str]:
        """
        Check whether the optimization has been running for too long.

        :return bool is_converged: True if the stopping criterion is fulfilled
        :return str terminate_reason: reason for the terminatiom
        """
        end_time = datetime.now()
        minutes_diff = (end_time - self.total_start_time).total_seconds() / 60.0
        terminate_reason = "Total optimization time exceeded limit."
        is_converged = False
        if minutes_diff >= self.total_time_lim:
            is_converged = True
            # this is a bit confusing but here we set is_running to False if is is linked
            if self.stop_opt_function is not None:
                self.stop_opt_function()
        return [is_converged, terminate_reason]



# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#  Copyright 2021-  QuOCS Team
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from quocslib.optimalcontrolproblems.su2 import hamiltonian_d1_d2
import numpy as np
from scipy.linalg import expm, norm
from quocslib.utils.AbstractFoM import AbstractFoM
import os


class OneQubit(AbstractFoM):
    def __init__(self, args_dict: dict = None):
        if args_dict is None:
            args_dict = {}

        self.psi_target = np.asarray(eval(args_dict.setdefault("target_state", "[1.0/np.sqrt(2), -1j/np.sqrt(2)]")),
                                     dtype="complex")
        self.psi_0 = np.asarray(eval(args_dict.setdefault("initial_state", "[1.0, 0.0]")), dtype="complex")
        self.delta1 = args_dict.setdefault("delta1", 0.1)
        self.delta2 = args_dict.setdefault("delta2", 0.1)

        # Noise in the figure of merit
        self.is_noisy = args_dict.setdefault("is_noisy", False)
        self.noise_factor = args_dict.setdefault("noise_factor", 0.05)
        self.std_factor = args_dict.setdefault("std_factor", 0.01)

        # Drifting FoM
        self.include_drift = args_dict.setdefault("include_drift", False)
        self.linear_drift_val_over_iterartion = args_dict.setdefault("linear_drift_val_over_iterartion", 0.002)

        # Maximization or minimization
        # Minimization -1.0
        # Maximization 1.0
        self.optimization_factor = args_dict.setdefault("optimization_factor", -1.0)

        self.FoM_list = []
        self.save_path = ""
        self.FoM_save_name = "FoM.txt"

        self.FoM_eval_number = 0

    # def __del__(self):
    #     np.savetxt(os.path.join(self.save_path, self.FoM_save_name), self.FoM_list)

    def save_FoM(self):
        np.savetxt(os.path.join(self.save_path, self.FoM_save_name), self.FoM_list)

    def set_save_path(self, save_path: str = ""):
        self.save_path = save_path

    def get_FoM(self, pulses: list = [], parameters: list = [], timegrids: list = []) -> dict:
        f = np.asarray(pulses[0])
        if parameters:
            # checks if list is empty, otherwise it sets delta1 to the first para
            self.delta1 = parameters[0]
        timegrid = np.asarray(timegrids[0])
        dt = timegrid[1] - timegrid[0]
        U = self._time_evolution(f, dt, self.delta1, self.delta2)
        psi_f = np.matmul(U, self.psi_0)
        infidelity = 1.0 - self._get_fidelity(self.psi_target, psi_f)
        std = 1e-4
        if self.is_noisy:
            noise = (self.noise_factor * 2 * (0.5 - np.random.rand(1)[0]))
            infidelity += noise
            std = (self.std_factor * np.random.rand(1)[0])

        if self.include_drift:
            infidelity += self.linear_drift_val_over_iterartion * self.FoM_eval_number

        self.FoM_list.append(np.abs(infidelity))
        self.FoM_eval_number += 1

        return {"FoM": (-1.0) * self.optimization_factor * np.abs(infidelity), "std": std}

    @staticmethod
    def _time_evolution(fc, dt, delta1, delta2):
        U = np.identity(2)
        for ii in range(fc.size - 1):
            ham_t = hamiltonian_d1_d2((fc[ii + 1] + fc[ii]) / 2, delta1=delta1, delta2=delta2)
            U_temp = U
            U = np.matmul(expm(-1j * ham_t * dt), U_temp)
        return U

    @staticmethod
    def _get_fidelity(psi1, psi2):
        return np.abs(np.dot(psi1.conj().T, psi2))**2 / (norm(psi1) * norm(psi2))
